#include "StdAfx.h"
#include "Grabber.h"

#include "Utility.h"

//=====================================================================================//
//                                    class Grabber                                    //
//=====================================================================================//
Grabber::Grabber(const std::string& file_name,unsigned int options)
: m_observer(0)
{
	setupGeneralSettings(file_name);

	if(options & Options::Video) 
		setupVideoGrabber();

	if(options & Options::Audio) 
		setupAudioGrabber();

	HRESULT hr = S_FALSE;

	//	query IMediaEvent interface
	if(FAILED(hr = m_graphBuilder.QueryInterface(&m_mediaEvent)))
		throw std::exception("failed to query IMediaEvent interface");

	//	query IMediaControl interface
	if(FAILED(hr = m_graphBuilder.QueryInterface(&m_mediaControl)))
		throw std::exception("failed to query IMediaControl interface");

	//	query IMediaSeeking interface
	if(FAILED(hr = m_graphBuilder.QueryInterface(&m_mediaSeeking)))
		throw std::exception("failed to query IMediaSeeking interface");

    // set the graph clock.
	ATL::CComPtr<IMediaFilter> media_filter;
    m_graphBuilder->QueryInterface(IID_IMediaFilter, (void**)&media_filter);
    media_filter->SetSyncSource(0);

	InitializeCriticalSection(&m_criticalSection);
}

Grabber::~Grabber()
{
	DeleteCriticalSection(&m_criticalSection);
}
//	setup general settings for graph builder
void Grabber::setupGeneralSettings(const std::string& file_name)
{
	HRESULT hr = S_FALSE;

	//	create graph builder
	if(FAILED(hr = m_graphBuilder.CoCreateInstance(CLSID_FilterGraph,NULL,CLSCTX_INPROC)))
		throw std::exception("can't create instance for IGraphBuilder interface");

	std::vector<wchar_t> buffer(file_name.size()+1);
	MultiByteToWideChar(CP_ACP,0,file_name.c_str(),-1,&buffer[0],buffer.size());

	//	add source filter to graph
	if(FAILED(hr = m_graphBuilder->AddSourceFilter(&buffer[0],L"Source",&m_sourceFilter)))
	{
		switch(hr)
		{
		case VFW_E_NOT_FOUND:
			throw std::exception(("file [" + file_name + "] not found").c_str());
		default:
			throw std::exception("can't add source filter to graph");
		}
	}

	//	create AVI splitter
	if(FAILED(hr = m_aviSplitter.CoCreateInstance(CLSID_AviSplitter,NULL,CLSCTX_INPROC_SERVER)))
		throw std::exception("can't create AVI Splitter filter");

	//	add AVI splitter to graph
	if(FAILED(hr = m_graphBuilder->AddFilter(m_aviSplitter,L"AVI Splitter")))
		throw std::exception("can't add AVI Splitter filter to graph");

	//	connect source filter to AVI splitter
	if(FAILED(hr = Utility::ConnectFilters(m_graphBuilder.p,m_sourceFilter.p,m_aviSplitter.p)))
		throw std::exception("can't connect source filter to AVI Splitter filter");
}
//	setup video grabber settings
void Grabber::setupVideoGrabber()
{
	HRESULT hr = S_FALSE;

	//	create video grabber
	ATL::CComPtr<IBaseFilter> video_sample_grabber_filter;
	if(FAILED(hr = video_sample_grabber_filter.CoCreateInstance(CLSID_SampleGrabber,NULL,CLSCTX_INPROC_SERVER)))
		throw std::exception("can't create instance for ISimpleGrabber interface");

	video_sample_grabber_filter.QueryInterface(&m_sampleVideoGrabber);

	AM_MEDIA_TYPE amt;
	ZeroMemory(&amt,sizeof(AM_MEDIA_TYPE));
	amt.majortype = MEDIATYPE_Video;
	amt.subtype = MEDIASUBTYPE_RGB24;
	if(FAILED(hr = m_sampleVideoGrabber->SetMediaType(&amt)))
		throw std::exception("can't set media type for video grabber filter");

	//	add video grabber to graph
	if(FAILED(hr = m_graphBuilder->AddFilter(video_sample_grabber_filter,L"Video Sample Grabber")))
		throw std::exception("can't add video grabber filter to graph");

	//	connect source filter to video grabber
	if(FAILED(hr = Utility::ConnectFilters(m_graphBuilder.p,m_aviSplitter.p,video_sample_grabber_filter.p)))
		throw std::exception("can't connect AVI Splitter filter to video grabber filter");

	//	create null renderer for video grabber filter
	if(FAILED(hr = m_nullVideoFilter.CoCreateInstance(CLSID_NullRenderer,0,CLSCTX_INPROC_SERVER)))
		throw std::exception("can't create instance for null renderer filter");

	//	add null renderer fliter to graph
	if(FAILED(hr = m_graphBuilder->AddFilter(m_nullVideoFilter,L"Null Video Renderer")))
		throw std::exception("can't add null renderer filter to graph");

	//	connect video grabber to null renderer
	if(FAILED(hr = Utility::ConnectFilters(m_graphBuilder.p,video_sample_grabber_filter.p,m_nullVideoFilter.p)))
		throw std::exception("can't connect grabber filter to renderer filter");

	m_videoGrabberCB = ATL::CComPtr<ISampleGrabberCB>(new VideoGrabber(*this));
}
//	setup audio grabber settings
void Grabber::setupAudioGrabber()
{
	HRESULT hr = S_FALSE;

	//	create audio grabber
	ATL::CComPtr<IBaseFilter> audio_sample_grabber_filter;
	if(FAILED(hr = audio_sample_grabber_filter.CoCreateInstance(CLSID_SampleGrabber,NULL,CLSCTX_INPROC_SERVER)))
		throw std::exception("can't create instance for ISimpleGrabber interface");

	audio_sample_grabber_filter.QueryInterface(&m_sampleAudioGrabber);

	AM_MEDIA_TYPE amt;
	ZeroMemory(&amt,sizeof(AM_MEDIA_TYPE));
	amt.formattype = FORMAT_WaveFormatEx;
	amt.majortype = MEDIATYPE_Audio;
	amt.subtype = MEDIASUBTYPE_PCM;
	if(FAILED(hr = m_sampleAudioGrabber->SetMediaType(&amt)))
		throw std::exception("can't set media type for audio grabber filter");

	//	add audio grabber to graph
	if(FAILED(hr = m_graphBuilder->AddFilter(audio_sample_grabber_filter,L"Sound Sample Grabber")))
		throw std::exception("can't add sample grabber filter to graph");

	//	connect source filter to audio grabber
	if(FAILED(hr = Utility::ConnectFilters(m_graphBuilder.p,m_aviSplitter.p,audio_sample_grabber_filter.p)))
		throw std::exception("can't connect AVI Splitter filter to sound grabber filter");

	//	create null renderer for audio grabber filter
	if(FAILED(hr = m_nullSoundFilter.CoCreateInstance(CLSID_NullRenderer,0,CLSCTX_INPROC_SERVER)))
		throw std::exception("can't create instance for null renderer filter");

	//	add null renderer fliter to graph
	if(FAILED(hr = m_graphBuilder->AddFilter(m_nullSoundFilter,L"Null Sound Renderer")))
		throw std::exception("can't add null renderer filter to graph");

	//	connect audio grabber to null renderer
	if(FAILED(hr = Utility::ConnectFilters(m_graphBuilder.p,audio_sample_grabber_filter.p,m_nullSoundFilter.p)))
		throw std::exception("can't connect grabber filter to renderer filter");

	m_audioGrabberCB = ATL::CComPtr<ISampleGrabberCB>(new AudioGrabber(*this));
}
//	run process
void Grabber::run(GrabberObserver* observer)
{
	HRESULT hr = E_FAIL;
	long code;

	m_observer = observer;

	if(m_sampleVideoGrabber)
	{
		m_sampleVideoGrabber->SetCallback(m_videoGrabberCB,0);
		m_sampleVideoGrabber->SetBufferSamples(TRUE);
	}

	if(m_sampleAudioGrabber)
	{
		m_sampleAudioGrabber->SetCallback(m_audioGrabberCB,0);
		m_sampleAudioGrabber->SetBufferSamples(TRUE);
	}

	hr = m_mediaControl->Run(); // Run the graph.
	hr = m_mediaEvent->WaitForCompletion(INFINITE,&code); // Wait till it's done.

	m_observer = 0;
}
//	process video data callback
void Grabber::onVideoData(const char* data,unsigned int size,unsigned int timestamp)
{
	EnterCriticalSection(&m_criticalSection);

	if(m_observer)
		m_observer->onVideoData(data,size,timestamp);

	LeaveCriticalSection(&m_criticalSection);
}
//	process audio data callback
void Grabber::onAudioData(const char* data,unsigned int size,unsigned int timestamp)
{
	EnterCriticalSection(&m_criticalSection);

	if(m_observer)
		m_observer->onAudioData(data,size,timestamp);

	LeaveCriticalSection(&m_criticalSection);
}
